@@ -87,8 +87,14 @@ char* ReadExifDate(char* File, int* IncludesGPS)
 {
 	// Open and read the file.
 	Exiv2::ExifData ExifRead;
+	Exiv2::Image::AutoPtr Image;
 
-	Exiv2::Image::AutoPtr Image = Exiv2::ImageFactory::open(File);
+	try {
+		Image = Exiv2::ImageFactory::open(File);
+	} catch (Exiv2::Error e) {
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
 	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
@@ -146,8 +152,14 @@ char* ReadExifData(char* File, double* Lat, double* Long, double* Elev, int* Inc
 	// for display purposes. For the GUI version.
 	// Open and read the file.
 	Exiv2::ExifData ExifRead;
+	Exiv2::Image::AutoPtr Image;
 
-	Exiv2::Image::AutoPtr Image = Exiv2::ImageFactory::open(File);
+	try {
+		Image = Exiv2::ImageFactory::open(File);
+	} catch (Exiv2::Error e) {
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
 	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
@@ -194,32 +206,26 @@ char* ReadExifData(char* File, double* Lat, double* Long, double* Elev, int* Inc
 		// Read it out and send it up!
 		// What we are trying to do here is convert the
 		// three rationals:
-		//    dd/1 mmmm/100 ss/1
+		//    dd/v mm/v ss/v
 		// To a decimal
 		//    dd.dddddd...
-		// dd/1 is easy: result = dd.
-		// mmmm/100 is harder:
-		//    mmmm
-		//    ---- / 60 = result.
-		//     100
-		// Combining a bit gives us:
-		//    mmmm / 6000 = decimal result
-		// Which is added to the degrees.
-		// ss/1 is sorta easy.
-		//     ss    1
-		//     -- * -- = result
-		//     60   60
-		// Combining gives us:
-		//     ss / 3600 = result.
-		// We only read ss to comply with the standard;
-		// we don't write ss as anything other than 0/1.
+		// dd/v is easy: result = dd/v.
+		// mm/v is harder:
+		//    mm
+		//    -- / 60 = result.
+		//     v
+		// ss/v is sorta easy.
+		//     ss   
+		//     -- / 3600 = result
+		//      v   
+		// Each part is added to the final number.
 		GPSData = ExifRead["Exif.GPSInfo.GPSLatitude"];
 		Exiv2::Rational RatNum = GPSData.toRational(0);
 		*Lat = (double)RatNum.first / (double)RatNum.second;
 		RatNum = GPSData.toRational(1);
-		*Lat = *Lat + ((double)RatNum.first / 6000);
+		*Lat = *Lat + (((double)RatNum.first / (double)RatNum.second) / 60);
 		RatNum = GPSData.toRational(2);
-		*Lat = *Lat + ((double)RatNum.first / 3600);
+		*Lat = *Lat + (((double)RatNum.first / (double)RatNum.second) / 3600);
 		
 		GPSData = ExifRead["Exif.GPSInfo.GPSLatitudeRef"];
 		if (strcmp(GPSData.toString().c_str(), "S") == 0)
@@ -232,9 +238,9 @@ char* ReadExifData(char* File, double* Lat, double* Long, double* Elev, int* Inc
 		RatNum = GPSData.toRational(0);
 		*Long = (double)RatNum.first / (double)RatNum.second;
 		RatNum = GPSData.toRational(1);
-		*Long = *Long + ((double)RatNum.first / 6000);
+		*Long = *Long + (((double)RatNum.first / (double)RatNum.second) / 60);
 		RatNum = GPSData.toRational(2);
-		*Long = *Long + ((double)RatNum.first / 3600);
+		*Long = *Long + (((double)RatNum.first / (double)RatNum.second) / 3600);
 		
 		GPSData = ExifRead["Exif.GPSInfo.GPSLongitudeRef"];
 		if (strcmp(GPSData.toString().c_str(), "W") == 0)
@@ -247,7 +253,6 @@ char* ReadExifData(char* File, double* Lat, double* Long, double* Elev, int* Inc
 		GPSData = ExifRead["Exif.GPSInfo.GPSAltitude"];
 		RatNum = GPSData.toRational(0);
 		*Elev = (double)RatNum.first / (double)RatNum.second;
-		
 	}
 
 
@@ -255,6 +260,86 @@ char* ReadExifData(char* File, double* Lat, double* Long, double* Elev, int* Inc
 	return Copy; // Its up to the caller to free this.
 };
 
+// This function is for the --fix-datestamp option.
+// DateStamp and TimeStamp should be 12-char strings.
+char* ReadGPSTimestamp(char* File, char* DateStamp, char* TimeStamp, int* IncludesGPS)
+{
+	// This function varies in that it reads
+	// much more data than the last, specifically
+	// for display purposes. For the GUI version.
+	// Open and read the file.
+	Exiv2::ExifData ExifRead;
+	Exiv2::Image::AutoPtr Image;
+
+	try {
+		Image = Exiv2::ImageFactory::open(File);
+	} catch (Exiv2::Error e) {
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
+	Image->readMetadata();
+	if (Image.get() == NULL)
+	{
+		//fprintf(stderr, "%s\n",
+		//	Exiv2::ExifData::strError(Result,
+		//		File).c_str());
+		fprintf(stderr, "Unable to open file %s.\n", File);
+		return 0;
+	}
+	
+	ExifRead = Image->exifData();
+
+	// Read the tag out.
+	Exiv2::Exifdatum& Tag = ExifRead["Exif.Photo.DateTimeOriginal"];
+
+	// Check that the tag is not blank.
+	std::string Value = Tag.toString();
+
+	if (Value.length() == 0)
+	{
+		// No date/time stamp.
+		// Not good.
+		// Just return - above us will handle it.
+		return 0;
+	}
+
+	// Copy the tag and return that.
+	char* Copy = (char*)malloc((sizeof(char) * Value.length()) + 1);
+	strcpy(Copy, Value.c_str());
+	
+	// Check if we have GPS tags.
+	Exiv2::Exifdatum& GPSData = ExifRead["Exif.GPSInfo.GPSVersionID"];
+
+	Value = GPSData.toString();
+
+	if (Value.length() == 0)
+	{
+		// No GPS data.
+		// Just return.
+		*IncludesGPS = 0;
+	} else {
+		// Seems to include GPS data...
+		*IncludesGPS = 1;
+		
+		// Read out the Time and Date stamp, for correction.
+		GPSData = ExifRead["Exif.GPSInfo.GPSTimeStamp"];
+		Exiv2::Rational RatNum1 = GPSData.toRational(0);
+		Exiv2::Rational RatNum2 = GPSData.toRational(1);
+		Exiv2::Rational RatNum3 = GPSData.toRational(2);
+		snprintf(TimeStamp, 12, "%02d:%02d:%02d",
+				RatNum1.first, RatNum2.first, RatNum3.first);
+		
+		GPSData = ExifRead["Exif.GPSInfo.GPSDateStamp"];
+		RatNum1 = GPSData.toRational(0);
+		RatNum2 = GPSData.toRational(1);
+		RatNum3 = GPSData.toRational(2);
+		snprintf(DateStamp, 12, "%04d:%02d:%02d",
+				RatNum1.first, RatNum2.first, RatNum3.first);
+	}
+
+	return Copy;
+};
+
 void ConvertToRational(double Number,
 			long int* Numerator, long int* Denominator,
 			int Rounding)
@@ -319,7 +404,7 @@ void ConvertToRational(double Number,
 
 }
 
-int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMtime)
+int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMtime, int DegMinSecs)
 {
 	// Write the GPS data to the file...
 
@@ -328,8 +413,14 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 	struct utimbuf utb;
 	if (NoChangeMtime)
 		stat(File, &statbuf);
+	Exiv2::Image::AutoPtr Image;
 
-	Exiv2::Image::AutoPtr Image = Exiv2::ImageFactory::open(File);
+	try {
+		Image = Exiv2::ImageFactory::open(File);
+	} catch (Exiv2::Error e) {
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
 	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
@@ -344,7 +435,8 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 
 	char ScratchBuf[100];
 	long int Nom, Denom;
-	long int Deg, Min;
+	long int Deg, Min, Sec;
+	double FracPart;
 
 	// Do all the easy constant ones first.
 	// GPSVersionID tag: standard says is should be four bytes: 02 00 00 00
@@ -358,13 +450,21 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 	// Now start adding data.
 	// ALTITUDE.
 	// Altitude reference: byte "00" meaning "sea level".
+	// Or "01" if the altitude value is negative.
 	Value = Exiv2::Value::create(Exiv2::unsignedByte);
-	Value->read("0");
+	if (Point->Elev > 0)
+	{
+		Value->read("0");
+	} else {
+		Value->read("1");
+	}
 	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSAltitudeRef"), Value.get());
 	// And the actual altitude.
-	Value = Exiv2::Value::create(Exiv2::signedRational);
+	Value = Exiv2::Value::create(Exiv2::unsignedRational);
 	ConvertToRational(Point->Elev, &Nom, &Denom, 4);
 	snprintf(ScratchBuf, 100, "%ld/%ld", Nom, Denom);
+
+	/* printf("Altitude: %f -> %s\n", Point->Elev, ScratchBuf); */
 	Value->read(ScratchBuf);
 	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSAltitude"), Value.get());
 	
@@ -381,6 +481,7 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 		ExifToWrite["Exif.GPSInfo.GPSLatitudeRef"] = "N";
 	}
 	// Now the actual lattitude itself.
+	// The original comment read:
 	// This is done as three rationals.
 	// I choose to do it as:
 	//   dd/1 - degrees.
@@ -395,10 +496,27 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 	// Further note: original code did not translate between
 	//   dd.dddddd to dd mm.mm - that's why we now multiply
 	//   by 6000 - x60 to get minutes, x100 to get to mmmm/100.
-	Value = Exiv2::Value::create(Exiv2::signedRational);
-	Deg = (int)floor(fabs(Point->Lat)); // Slice off after decimal.
-	Min = (int)floor((fabs(Point->Lat) - floor(fabs(Point->Lat))) * 6000);
-	snprintf(ScratchBuf, 100, "%ld/1 %ld/100 0/1", Deg, Min);
+	//
+	// Rereading the EXIF standard, it's quite ok to do DD MM SS.SS
+	// Which is much more accurate. This is the new default, unless otherwise
+	// set.
+	Value = Exiv2::Value::create(Exiv2::unsignedRational);
+
+	if (DegMinSecs)
+	{
+		Deg = (int)floor(fabs(Point->Lat)); // Slice off after decimal.
+		Min = (int)floor((fabs(Point->Lat) - floor(fabs(Point->Lat))) * 60); // Now grab just the minutes.
+		FracPart = ((fabs(Point->Lat) - floor(fabs(Point->Lat))) * 60) - (double)Min; // Grab the fractional minute.
+		Sec = (int)floor(FracPart * 6000); // Convert to seconds.
+		
+		/* printf("New style lattitude: %f -> %ld/%ld/ %ld/100\n", Point->Lat, Deg, Min, Sec); */
+
+		snprintf(ScratchBuf, 100, "%ld/1 %ld/1 %ld/100", Deg, Min, Sec);
+	} else {
+		Deg = (int)floor(fabs(Point->Lat)); // Slice off after decimal.
+		Min = (int)floor((fabs(Point->Lat) - floor(fabs(Point->Lat))) * 6000);
+		snprintf(ScratchBuf, 100, "%ld/1 %ld/100 0/1", Deg, Min);
+	}
 	Value->read(ScratchBuf);
 	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSLatitude"), Value.get());
 	
@@ -429,10 +547,23 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 	// Further note: original code did not translate between
 	//   dd.dddddd to dd mm.mm - that's why we now multiply
 	//   by 6000 - x60 to get minutes, x100 to get to mmmm/100.
-	Value = Exiv2::Value::create(Exiv2::signedRational);
-	Deg = (int)floor(fabs(Point->Long)); // Slice off after decimal.
-	Min = (int)floor((fabs(Point->Long) - floor(fabs(Point->Long))) * 6000);
-	snprintf(ScratchBuf, 100, "%ld/1 %ld/100 0/1", Deg, Min);
+	Value = Exiv2::Value::create(Exiv2::unsignedRational);
+
+	if (DegMinSecs)
+	{
+		Deg = (int)floor(fabs(Point->Long)); // Slice off after decimal.
+		Min = (int)floor((fabs(Point->Long) - floor(fabs(Point->Long))) * 60); // Now grab just the minutes.
+		FracPart = ((fabs(Point->Long) - floor(fabs(Point->Long))) * 60) - (double)Min; // Grab the fractional minute.
+		Sec = (int)floor(FracPart * 6000); // Convert to seconds.
+
+		/* printf("New style longitude: %f -> %ld/%ld/ %ld/100\n", Point->Long, Deg, Min, Sec); */
+
+		snprintf(ScratchBuf, 100, "%ld/1 %ld/1 %ld/100", Deg, Min, Sec);
+	} else {
+		Deg = (int)floor(fabs(Point->Long)); // Slice off after decimal.
+		Min = (int)floor((fabs(Point->Long) - floor(fabs(Point->Long))) * 6000);
+		snprintf(ScratchBuf, 100, "%ld/1 %ld/100 0/1", Deg, Min);
+	}
 	Value->read(ScratchBuf);
 	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSLongitude"), Value.get());
 
@@ -441,7 +572,19 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 	// The timestamp is taken as the UTC time of the photo.
 	// If interpolation occured, then this time is the time of the photo.
 	struct tm TimeStamp;
+	TimeStamp.tm_isdst = -1;
 	gmtime_r(&(Point->Time), &TimeStamp);
+	TimeStamp.tm_isdst = -1;
+
+	if (Point->Time != mktime(&TimeStamp)) {
+		// What happened is gmtime subtracted the current time zone.
+		// I thought it was called "gmtime" for a reason.
+		// Oh well. Add the difference and try again.
+		// This is a hack.
+		time_t CorrectedTime = Point->Time + (Point->Time - mktime(&TimeStamp));
+
+		gmtime_r(&CorrectedTime, &TimeStamp);
+	}
 
 	Value = Exiv2::Value::create(Exiv2::signedRational);
 	snprintf(ScratchBuf, 100, "%d/1 %d/1 %d/1",
@@ -474,6 +617,81 @@ int WriteGPSData(char* File, struct GPSPoint* Point, char* Datum, int NoChangeMt
 	
 };
 
+int WriteFixedDatestamp(char* File, time_t Time)
+{
+	// Write the GPS data to the file...
+
+	struct stat statbuf;
+	struct stat statbuf2;
+	struct utimbuf utb;
+	stat(File, &statbuf);
+
+	Exiv2::Image::AutoPtr Image;
+
+	try {
+		Image = Exiv2::ImageFactory::open(File);
+	} catch (Exiv2::Error e) {
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
+	Image->readMetadata();
+	if (Image.get() == NULL)
+	{
+		// It failed if we got here.
+		//fprintf(stderr, "%s\n",
+		//	Exiv2::ExifData::strError(Result, File).c_str());
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
+	
+	Exiv2::ExifData &ExifToWrite = Image->exifData();
+	
+	struct tm TimeStamp;
+	TimeStamp.tm_isdst = -1;
+	gmtime_r(&Time, &TimeStamp);
+	TimeStamp.tm_isdst = -1;
+
+	if (Time != mktime(&TimeStamp)) {
+		// What happened is gmtime subtracted the current time zone.
+		// I thought it was called "gmtime" for a reason.
+		// Oh well. Add the difference and try again.
+		// This is a hack.
+		time_t CorrectedTime = Time + (Time - mktime(&TimeStamp));
+
+		gmtime_r(&CorrectedTime, &TimeStamp);
+	}
+
+	char ScratchBuf[100];
+
+	Exiv2::Value::AutoPtr Value;
+	Value = Exiv2::Value::create(Exiv2::signedRational);
+	snprintf(ScratchBuf, 100, "%d/1 %d/1 %d/1",
+			TimeStamp.tm_year + 1900,
+			TimeStamp.tm_mon + 1,
+			TimeStamp.tm_mday);
+	Value->read(ScratchBuf);
+	ExifToWrite.erase(ExifToWrite.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSDateStamp")));
+	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSDateStamp"), Value.get());
+	
+	Value = Exiv2::Value::create(Exiv2::signedRational);
+	snprintf(ScratchBuf, 100, "%d/1 %d/1 %d/1",
+			TimeStamp.tm_hour, TimeStamp.tm_min,
+			TimeStamp.tm_sec);
+	Value->read(ScratchBuf);
+	ExifToWrite.erase(ExifToWrite.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSTimeStamp")));
+	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSTimeStamp"), Value.get());
+	
+	Image->writeMetadata();
+	
+	// Reset the mtime.
+	stat(File, &statbuf2);
+	utb.actime = statbuf2.st_atime;
+	utb.modtime = statbuf.st_mtime;
+	utime(File, &utb);
+
+	return 1;
+}
+
 int RemoveGPSExif(char* File, int NoChangeMtime)
 {
 	struct stat statbuf;
@@ -483,8 +701,14 @@ int RemoveGPSExif(char* File, int NoChangeMtime)
 		stat(File, &statbuf);
 
 	// Open the file and start reading.
+	Exiv2::Image::AutoPtr Image;
 	
-	Exiv2::Image::AutoPtr Image = Exiv2::ImageFactory::open(File);
+	try {
+		Image = Exiv2::ImageFactory::open(File);
+	} catch (Exiv2::Error e) {
+		fprintf(stderr, "Failed to open file %s.\n", File);
+		return 0;
+	}
 	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
@@ -497,25 +721,17 @@ int RemoveGPSExif(char* File, int NoChangeMtime)
 
 	Exiv2::ExifData &ExifInfo = Image->exifData();
 
-	Exiv2::ExifData::iterator Iter;
-
 	// Search through, find the keys that we want, and wipe them
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSVersionID"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSLatitudeRef"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSLatitude"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSLongitudeRef"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSLongitude"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSAltitudeRef"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSAltitude"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
-	Iter = ExifInfo.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSTimeStamp"));
-	if (Iter != ExifInfo.end()) ExifInfo.erase(Iter);
+	// Code below submitted by Marc Horowitz
+	Exiv2::ExifData::iterator Iter;
+	for (Exiv2::ExifData::iterator Iter = ExifInfo.begin();
+		Iter != ExifInfo.end(); )
+	{
+		if (Iter->key().find("Exif.GPSInfo") == 0)
+			Iter = ExifInfo.erase(Iter);
+		else
+			Iter++;
+	}
 	
 	Image->writeMetadata();
 
