@@ -102,7 +102,7 @@ void ExtractTrackPoints(xmlNodePtr Start)
 
 			/* Check that we have all the data. If we're missing something,
 			 * then skip this point... */
-			if (Time == NULL || Elev == NULL || Long == NULL || Lat == NULL)
+			if (Time == NULL || Long == NULL || Lat == NULL)
 			{
 				/* Missing some data. */
 				/* TODO: Really should report this upstream... */
@@ -133,7 +133,9 @@ void ExtractTrackPoints(xmlNodePtr Start)
 			/* Write the data into LastPoint, which should be a new point. */
 			LastPoint->Lat = atof(Lat);
 			LastPoint->Long = atof(Long);
-			LastPoint->Elev = atof(Elev);
+			if (Elev) {
+				LastPoint->Elev = atof(Elev);
+			}
 			LastPoint->Time = ConvertToUnixTime(Time, GPX_DATE_FORMAT, 0, 0);
 			
 			/* Debug...
