@@ -54,9 +54,19 @@ static struct option program_options[] = {
 	{ "remove", no_argument, 0, 'r'},
 	{ "ignore-tracksegs", no_argument, 0, 't'},
 	{ "no-mtime", no_argument, 0, 'M'},
+	{ "version", no_argument, 0, 'V'},
+	{ "fix-datestamps", no_argument, 0, 'f'},
+	{ "degmins", no_argument, 0, 'p'},
+	{ "photooffset", required_argument, 0, 'O'},
 	{ 0, 0, 0, 0 }
 };
 
+/* Function to print the version - near the top for easy modification. */
+void PrintVersion(char* ProgramName)
+{
+	printf("%s, v 1.5.4, 22 July 2007. Daniel Foote, 2005-2007. GNU GPL.\n", ProgramName);
+}
+
 /* Function to print the usage info. */
 void PrintUsage(char* ProgramName)
 {
@@ -77,7 +87,11 @@ void PrintUsage(char* ProgramName)
 	printf("--remove or -r: Strip GPS tags from the given files, and then quit.\n");
 	printf("--ignore-tracksegs or -t: Interpolate between track segments too.\n");
 	printf("--no-mtime or -M: Don't change mtime of modified files.\n");
+	printf("--fix-datestamps or -f: Fix broken GPS datestamps written with versions < 1.5.2\n");
+	printf("--degmins or -p: Write location as DD MM.MM as was default before < 1.5.3.\n");
+	printf("--photooffset or -O <seconds>: Offset added to photo time to make it match the GPS.\n");
 	printf("--help or -h: display usage/help message.\n");
+	printf("--version or -V: display version information.\n");
 }
 
 /* Display the information from an existing file. */
@@ -137,6 +151,59 @@ void RemoveGPSTags(char* File, int NoChangeMtime)
 	}
 }
 
+/* Fix GPSDatestamp tags, if they were incorrect, as found with versions
+ * earlier than 1.5.2. */
+void FixDatestamp(char* File, int AdjustmentHours, int AdjustmentMinutes, int NoWriteExif)
+{
+	/* Read the timestamp data. */
+	char DateStamp[12];
+	char TimeStamp[12];
+	char CombinedTime[24];
+	int IncludesGPS = 0;
+	char* OriginalDateStamp = NULL;
+
+	OriginalDateStamp = ReadGPSTimestamp(File, DateStamp, TimeStamp, &IncludesGPS);
+
+	if (OriginalDateStamp == NULL)
+	{
+		printf("%s: No EXIF data.\n", File);
+	} else if (IncludesGPS == 0) {
+		printf("%s: No GPS data.\n", File);
+	} else {
+		/* Check the timestamp. */
+		time_t PhotoTime = ConvertToUnixTime(OriginalDateStamp, EXIF_DATE_FORMAT,
+			AdjustmentHours, AdjustmentMinutes);
+		
+		strcpy(CombinedTime, DateStamp);
+		strcat(CombinedTime, " ");
+		strcat(CombinedTime, TimeStamp);
+
+		time_t GPSTime = ConvertToUnixTime(CombinedTime, EXIF_DATE_FORMAT, 0, 0);
+
+		if (PhotoTime != GPSTime) {
+			/* Timestamp is wrong. Fix it.
+			 * Should be photo time - this also corrects
+			 * GPSTimestamp, which was wrong too. */
+			if (!NoWriteExif)
+			{
+				WriteFixedDatestamp(File, PhotoTime);
+			}
+			char PhotoTimeFormat[100];
+			char GPSTimeFormat[100];
+			ctime_r(&PhotoTime, PhotoTimeFormat);
+			ctime_r(&GPSTime, GPSTimeFormat);
+			printf("%s: Wrong timestamp:\n   Photo:     %s   GPS:       %s   Corrected: %s", 
+					File, PhotoTimeFormat, GPSTimeFormat, PhotoTimeFormat);
+		} else {
+			/* Inside the range. Do nothing! */
+			printf("%s: Timestamp is OK: Photo %s (localtime), GPS %s (UTC).\n", File,
+					OriginalDateStamp, CombinedTime);
+		}
+	}
+
+	free(OriginalDateStamp);
+}
+
 int main(int argc, char** argv)
 {
 	/* Say hello. */
@@ -168,12 +235,15 @@ int main(int argc, char** argv)
 	int RemoveTags = 0;
 	int DoBetweenTrackSegs = 0;
 	int NoChangeMtime = 0;
+	int FixDatestamps = 0;
+	int DegMinSecs = 1;
+	int PhotoOffset = 0;
 
 	while (1)
 	{
 		/* Call getopt to do all the hard work
 		 * for us... */
-		c = getopt_long(argc, argv, "g:z:ihvd:m:nsortM",
+		c = getopt_long(argc, argv, "g:z:ihvd:m:nsortMVfO",
 				program_options, 0);
 
 		if (c == -1) break;
@@ -201,6 +271,9 @@ int main(int argc, char** argv)
 					strncpy(TimeAdjustment, optarg, strlen(optarg)+1);
 				}
 				break;
+			case 'O':
+				PhotoOffset = atoi(optarg);
+				break;
 			case 'i':
 				/* This option disables interpolation. */
 				Interpolate = 0;
@@ -230,6 +303,15 @@ int main(int argc, char** argv)
 				PrintUsage(argv[0]);
 				exit(EXIT_SUCCESS);
 				break;
+			case 'V':
+				/* Display version information, and then quit. */
+				PrintVersion(argv[0]);
+				exit(EXIT_SUCCESS);
+				break;
+			case 'f':
+				/* Fix Datestamps. */
+				FixDatestamps = 1;
+				break;
 			case 's':
 				/* Show the data in the photos. Mark this. */
 				ShowOnlyDetails = 1;
@@ -250,6 +332,10 @@ int main(int argc, char** argv)
 			case 'M':
 				NoChangeMtime = 1;
 				break;
+			case 'p':
+				/* Write in old DegMins format. */
+				DegMinSecs = 0;
+				break;
 			case '?':
 				/* Unrecognised option. Or, missing argument. */
 				/* We should inform the user and let them correct this. */
@@ -294,6 +380,34 @@ int main(int argc, char** argv)
 		exit(EXIT_SUCCESS);
 	}
 
+	/* If we wanted to fix datestamps, do this now.
+	 * This is to fix incorrect GPSDateStamp values written by versions
+	 * earlier than 1.5.2. */
+	if (FixDatestamps)
+	{
+		if (TimeAdjustment == NULL)
+		{
+			printf("You must give a time adjustment for the photos with -z to fix photos.\n");
+			exit(EXIT_SUCCESS);
+		}
+		
+		/* Break up the adjustment and convert to numbers. */
+		if (strstr(TimeAdjustment, ":"))
+		{
+			/* Found colon. Split into two. */
+			sscanf(TimeAdjustment, "%d:%d", &TimeZoneHours, &TimeZoneMins);
+		} else {
+			/* No colon. Just parse. */
+			TimeZoneHours = atoi(TimeAdjustment);
+		}
+
+		while (optind < argc)
+		{
+			FixDatestamp(argv[optind++], TimeZoneHours, TimeZoneMins, NoWriteExif);
+		}
+		exit(EXIT_SUCCESS);
+	}
+
 	/* Set up any other command line options... */
 	if (!Datum)
 	{
@@ -360,6 +474,8 @@ int main(int argc, char** argv)
 	Options.Datum         = Datum;
 	Options.DoBetweenTrkSeg = DoBetweenTrackSegs;
 	Options.NoChangeMtime = NoChangeMtime;
+	Options.DegMinSecs    = DegMinSecs;
+	Options.PhotoOffset   = PhotoOffset;
 
 	Options.MinTime = 0;
 	Options.MaxTime = 0;
